@@ -26,17 +26,93 @@
 #include <NetworkManagerSettingsInterface.h>
 #include <NetworkManagerSettingsConnectionInterface.h>
 #include <URfkillInterface.h>
+#include <util/localisation.h>
 
 #include <QStringList>
 #include <QDBusReply>
 #include <QtDebug>
 #include <QDBusInterface>
 #include <QDBusMetaType>
+#include <QRandomGenerator>
 #include <QRegularExpression>
 #include <NetworkManager.h>
 
+#ifndef DEFAULT_PASSWORD_LEN
+#define DEFAULT_PASSWORD_LEN 12
+#endif // !DEFAULT_PASSWORD_LEN
+
 using namespace std;
 
+namespace {
+    // TODO: barely any of these codes are used. Do we really need all of these?
+    QString reasonToString (int reason) {
+        switch (reason) {
+            case 0:
+                return _("Unknown error");
+            case 1:
+                return _("No reason given");
+            case 2:
+                return _("Device is now managed");
+            case 3:
+                return _("Device is now unmanaged");
+            case 4:
+                return _("The device could not be readied for configuration");
+            case 5:
+                return _("IP configuration could not be reserved (no available address, timeout, etc.)");
+            case 6:
+                return _("The IP configuration is no longer valid");
+            case 7:
+                return _("Your authentication details were incorrect");
+            case 8:
+                return _("802.1X supplicant disconnected");
+            case 9:
+                return _("802.1X supplicant configuration failed");
+            case 10:
+                return _("802.1X supplicant failed");
+            case 11:
+                return _("802.1X supplicant took too long to authenticate");
+            case 15:
+                return _("DHCP client failed to start");
+            case 16:
+                return _("DHCP client error");
+            case 17:
+                return _("DHCP client failed");
+            case 18:
+                return _("Shared connection service failed to start");
+            case 19:
+                return _("Shared connection service failed");
+            case 35:
+                return _("Necessary firmware for the device may be missing");
+            case 36:
+                return _("The device was removed");
+            case 37:
+                return _("NetworkManager went to sleep");
+            case 38:
+                return _("The device's active connection disappeared");
+            case 39:
+                return _("Device disconnected by user or client");
+            case 41:
+                return _("The device's existing connection was assumed");
+            case 42:
+                return _("The supplicant is now available");
+            case 43:
+                return _("The modem could not be found");
+            case 44:
+                return _("The Bluetooth connection failed or timed out");
+            case 50:
+                return _("A dependency of the connection failed");
+            case 52:
+                return _("ModemManager is unavailable");
+            case 53:
+                return _("The Wi-Fi network could not be found");
+            case 54:
+                return _("A secondary connection of the base connection failed");
+            default:
+                return _("Unknown");
+        }
+    }
+}
+
 namespace nmofono
 {
 
@@ -126,10 +202,11 @@ public:
                     m_manager->connection());
 
         int count = 0;
-        // Wait for connection to activate
-        while (count < 20 && activeConnection.state() != NM_ACTIVE_CONNECTION_STATE_ACTIVATED)
+        // Wait for connection to activate. Waits 100 ms x 100 = 10 secs.
+        while (count < 100 && activeConnection.state() != NM_ACTIVE_CONNECTION_STATE_ACTIVATED)
         {
-            qDebug() << "Waiting for hotspot to connect";
+            if (count % 10 == 0) // Only log once every second
+                qDebug() << "Waiting for hotspot to connect";
             QThread::msleep(100);
             ++count;
         }
@@ -519,16 +596,17 @@ public:
 
     void generatePassword()
     {
-        static const std::string items("abcdefghijklmnopqrstuvwxyz01234567890");
-        const int password_length = 8;
-        std::string result;
+        static const QString symbols = QLatin1String("01234567890"
+            "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz");
+        QString result;
+        QRandomGenerator *grng = QRandomGenerator::global();
 
-        for (int i = 0; i < password_length; i++)
+        for (int i = 0; i < DEFAULT_PASSWORD_LEN; i++)
         {
-            result.push_back(items[std::rand() % items.length()]);
+            result.push_back(symbols[grng->bounded(symbols.length())]);
         }
 
-        m_password = QString::fromStdString(result);
+        m_password = result;
     }
 
     void setDisconnectWifi(bool disconnect)
@@ -571,6 +649,16 @@ public Q_SLOTS:
         }
     }
 
+    void sendErrorNotification(int reason)
+    {
+        const QString summary = _("Failed to enable hotspot");
+        const QString message = reasonToString(reason);
+        const QString icon = QStringLiteral("hotspot-disabled");
+
+        m_notificationManager->notify(summary, message, icon,
+            /* actions */ {}, /* hints */ {}, /* timeout */ 5)->show();
+    }
+
 public:
     HotspotManager& p;
 
@@ -609,14 +697,18 @@ public:
     QString m_uuid;
 
     connection::ActiveConnectionManager::SPtr m_activeConnectionManager;
+
+    notify::NotificationManager::SPtr m_notificationManager;
 };
 
 HotspotManager::HotspotManager(connection::ActiveConnectionManager::SPtr activeConnectionManager,
+                               notify::NotificationManager::SPtr notificationManager,
                                const QDBusConnection& connection,
                                QObject *parent) :
         QObject(parent), d(new Priv(*this))
 {
     d->m_activeConnectionManager = activeConnectionManager;
+    d->m_notificationManager = notificationManager;
 
     d->m_manager = make_unique<OrgFreedesktopNetworkManagerInterface>(
             NM_DBUS_SERVICE, NM_DBUS_PATH, connection);
@@ -635,6 +727,8 @@ HotspotManager::HotspotManager(connection::ActiveConnectionManager::SPtr activeC
     {
         d->updateSettingsFromDbus();
     }
+
+    QObject::connect(this, &HotspotManager::reportError, d.get(), &Priv::sendErrorNotification);
 }
 
 void HotspotManager::setEnabled(bool value)
@@ -681,14 +775,22 @@ void HotspotManager::setEnabled(bool value)
         }
         d->enable(d->m_device->m_path);
     }
-    else
+
+    // We disable the hotspot if requested to be disabled, or if requested to be
+    // enabled but we fail to enable it. We can do this as d->enable() is fully
+    // synchronous (sigh).
+    if (!value || (value && !enabled()))
     {
-        // Disabling the hotspot.
         d->disable();
 
         setMtkTetheringEnabled(false);
 
         d->setDisconnectWifi(false);
+
+        if (value) {
+            // FIXME: is this a good code?
+            Q_EMIT reportError(NM_DEVICE_STATE_REASON_CONFIG_FAILED);
+        }
     }
 
 }
